# frozen_string_literal: true

require 'httparty'
require 'open3'

# Distributed Press (https://distributed.press) API client
class DistributedPress
  include ::HTTParty

  NEWLINE  = "\r\n"
  BOUNDARY = '--'

  attr_reader :api_key, :project_domain

  # @param options [Hash]
  # @option options [String] :url API URL
  # @option options [String] :api_key API Key
  # @option options [String] :project_domain Domain name
  def initialize(url: 'https://api.distributed.press', api_key: nil, project_domain: nil)
    @api_key = api_key || ENV['DISTRIBUTED_PRESS_API_KEY']
    @project_domain = project_domain || ENV['DISTRIBUTED_PRESS_PROJECT_DOMAIN']

    raise ArgumentError, "API key is missing" unless @api_key
    raise ArgumentError, "Project domain is missing" unless @project_domain

    self.class.default_options[:base_uri] = HTTParty.normalize_base_uri(url)
  end

  # Configure the domain
  #
  # @return [Boolean] Configuration was successful
  def configure
    stream, config = IO.pipe

    Thread.new do
      write_multipart_header config, 'config.json'
      IO.copy_stream StringIO.new({ 'domain' => project_domain }.to_json), config
      write_multipart_footer config

      config.close
    end

    result = upload_io('/v0/publication/configure', stream)

    result['errorCode'].zero?
  end

  # Publishes a directory by tarballing it on the fly.
  #
  # @param [String] Directory path
  # @return [Boolean] Upload was successful
  def publish(path)
    raise ArgumentError, "#{path} is not a directory" unless File.directory? path

    filename = File.basename path

    Open3.popen2('tar', '--to-stdout', '--create', '--gzip', '--dereference', '--directory', path, '.') do |stdin, stdout, thread|
      stream, body = IO.pipe

      # Generate a multipart body and copy contents to it.
      Thread.new do
        write_multipart_header body, filename
        IO.copy_stream stdout, body
        write_multipart_footer body

        body.close
      end

      result = upload_io '/v0/publication/publish', stream

      # wait
      thread.value
      result['errorCode'].zero?
    end
  end

  private

  # Write the multipart header
  def write_multipart_header(io, filename)
    io.write BOUNDARY
    io.write boundary
    io.write NEWLINE
    io.write "Content-Disposition: form-data; name=\"file\"; filename=\"#{filename}.tar.gz\""
    io.write NEWLINE
    io.write 'Content-Type: application/octet-stream'
    io.write NEWLINE
    io.write NEWLINE

    nil
  end

  # Write the multipart footer
  def write_multipart_footer(io)
    io.write NEWLINE
    io.write BOUNDARY
    io.write boundary
    io.write BOUNDARY
    io.write NEWLINE

    nil
  end

  # Upload an IO object
  #
  # @param [String]
  # @param [IO]
  def upload_io(endpoint, io)
    self.class.post(endpoint, body_stream: io, headers: headers(
      'Accept' => 'application/json',
      'Content-Type' => "multipart/form-data; boundary=#{boundary}",
      'Transfer-Encoding' => 'chunked'
    ))
  end

  # Generate a MultiPart boundary and reuse it
  #
  # @return [String]
  def boundary
    @boundary ||= HTTParty::Request::MultipartBoundary.generate
  end

  # Default request headers
  #
  # @param [Hash] Extra headers
  # @return [Hash]
  def headers(extra = {})
    extra.merge('Accept' => 'application/json', 'Authorization' => 'Bearer ' + api_key)
  end
end
